import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.geom.*;
import javax.imageio.*;
import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;

enum NodeType {
	START, END, JUNCTION, DEADEND, CORNER, WALL;
}

enum Direction {
	NORTH(0), EAST(1), SOUTH(2), WEST(3);
	
	private final int value;
	
	Direction(final int value) {
		this.value = value;
	}
	
	public int value() {
		return value;
	}
}

class Node {
	public NodeType type;
	private int x, y;
	private Node[] connections;
	boolean visited = false;
	
	public Node(NodeType type, int x, int y, Node[] connections) {
		this.type = type;
		setCoords(x, y);
		setConnections(connections);
	}
	
	public void setX(int x) throws IllegalArgumentException {
		if (x >= 0) {
			this.x = x;
		}
		else {
			throw new IllegalArgumentException("x must be positive");
		}
	}
	
	public void setY(int y) throws IllegalArgumentException {
		if (y >= 0) {
			this.y = y;
		}
		else {
			throw new IllegalArgumentException("y must be positive");
		}
	}
	
	public void setCoords(int x, int y) {
		setX(x);
		setY(y);
	}
	
	public void setConnections(Node[] connections) {
		this.connections = connections;
	}
	
	public void setConnection(Direction direction, Node connection) {
		connections[direction.value()] = connection;
	}
	
	public int getX() {
		return x;
	}
	
	public int getY() {
		return y;
	}
	
	public int[] getCoords() {
		return new int[]{x, y};
	}
	
	public Node[] getConnections() {
		return connections;
	}
	
	public Node getConnection(Direction direction) {
		return connections[direction.value()];
	}
	
	public boolean isStart() {
		return type == NodeType.START;
	}
	public boolean isEnd() {
		return type == NodeType.END;
	}
	public boolean isCorner() {
		return type == NodeType.CORNER;
	}
	public boolean isJunction() {
		return type == NodeType.JUNCTION;
	}
	public boolean isDeadEnd() {
		return type == NodeType.DEADEND;
	}
	public boolean isWall() {
		return type == NodeType.WALL;
	}
}



class Maze {
	public static final String ANSI_RESET = "\u001B[0m";
	public static final String ANSI_BLACK = "\u001B[30m";
	public static final String ANSI_GREEN_BG = "\u001B[42m";
	public static final String ANSI_BLUE_BG = "\u001B[44m";
	public static final String ANSI_BRIGHT_BLACK_BG = "\u001B[100m";
	public static final String ANSI_BRIGHT_WHITE_BG = "\u001B[107m";
	public static final int BLUE = -16776961;
	public static final int GREEN = -16711936;
	public static final int BLACK = -16777216
;
	
	private BufferedImage maze;
	
	public Maze(BufferedImage maze) {
		this.maze = maze;
	}
	
	private int[][] imageToArray() {
		//https://stackoverflow.com/questions/6524196/java-get-pixel-array-from-image
		final byte[] pixels = ((DataBufferByte) maze.getRaster().getDataBuffer()).getData();
		final int width = maze.getWidth();
		final int height = maze.getHeight();
		final boolean hasAlphaChannel = maze.getAlphaRaster() != null;

		int[][] result = new int[height][width];
		if (hasAlphaChannel) {
			final int pixelLength = 4;
			for (int pixel = 0, row = 0, col = 0; pixel + 3 < pixels.length; pixel += pixelLength) {
				int argb = 0;
				argb += (((int) pixels[pixel] & 0xff) << 24); // alpha
				argb += ((int) pixels[pixel + 1] & 0xff); // blue
				argb += (((int) pixels[pixel + 2] & 0xff) << 8); // green
				argb += (((int) pixels[pixel + 3] & 0xff) << 16); // red
				result[row][col] = argb;
				col++;
				if (col == width) {
					col = 0;
					row++;
				}
			}
		} else {
			final int pixelLength = 3;
			for (int pixel = 0, row = 0, col = 0; pixel + 2 < pixels.length; pixel += pixelLength) {
				int argb = 0;
				argb += BLACK; // 255 alpha
				argb += ((int) pixels[pixel] & 0xff); // blue
				argb += (((int) pixels[pixel + 1] & 0xff) << 8); // green
				argb += (((int) pixels[pixel + 2] & 0xff) << 16); // red
				result[row][col] = argb;
				col++;
				if (col == width) {
					col = 0;
					row++;
				}
			}
		}
		return result;
	}
	
	private Node[][] parse() {
		final int width = maze.getWidth();
		final int height = maze.getHeight();
		final int[][] mazeArray = imageToArray();
		Node[][] nodes = new Node[height][width];
		
		for (int x = 0; x < height; x++) {
			for (int y = 0; y < width; y++) {
				if (mazeArray[x][y] == BLUE) {
					nodes[x][y] = new Node(NodeType.START, y, x, null);
				} else if (mazeArray[x][y] == GREEN) {
					nodes[x][y] = new Node(NodeType.END, y, x, null);
				} else if (mazeArray[x][y] == BLACK) {
					nodes[x][y] = new Node(NodeType.WALL, y, x, null);
				} else {
					boolean[] connections = new boolean[4];
					connections[0] = (x > 0) ? (mazeArray[x-1][y] == -1 || mazeArray[x-1][y] == BLUE || mazeArray[x-1][y] == GREEN) : false;
					connections[1] = (y < height-1) ? (mazeArray[x][y+1] == -1 || mazeArray[x][y+1] == BLUE || mazeArray[x][y+1] == GREEN) : false;
					connections[2] = (x < width-1) ? (mazeArray[x+1][y] == -1 || mazeArray[x+1][y] == BLUE || mazeArray[x+1][y] == GREEN) : false;
					connections[3] = (y > 0) ? (mazeArray[x][y-1] == -1 || mazeArray[x][y-1] == BLUE || mazeArray[x][y-1] == GREEN) : false;
					
					int connectionCount = 0;
					int directionSum = 0;
					for (int i = 0; i < 4; i++) {
						if (connections[i]) {
							directionSum += i;
							connectionCount++;
						}
					}
					
					if (connectionCount == 1) {
						nodes[x][y] = new Node(NodeType.DEADEND, y, x, null);
					} else if (connectionCount == 2 && directionSum % 2 == 1) {
						nodes[x][y] = new Node(NodeType.CORNER, y, x, null);
					} else if (connectionCount > 2) {
						nodes[x][y] = new Node(NodeType.JUNCTION, y, x, null);
					}
				}
			}
		}
		return nodes;
	}
	
	private Node reclassifyNode(Node node, Node[] connections) {
		Node[] validConnections = new Node[4];
		int validConnectionCount = 0;
		int directionSum = 0;
		for (int i = 0; i < 4; i++) {
			Node connection = connections[i];
			if (connection != null) {
				if (!(connection.isDeadEnd() || connection.isWall()) && !connection.visited) {
					validConnections[i] = connections[i];
					validConnectionCount++;
					directionSum += i;
				} else if (!(connection.isDeadEnd() || connection.isWall()) && connection.visited) {
					directionSum += i;
					connections[i] = null;
				} else {
					connections[i] = null;
				}
			}
		}
		
		if (!node.isStart() && !node.isEnd()) {
			if (validConnectionCount == 0) {
				node.type = NodeType.DEADEND;
			} else if (validConnectionCount == 1 && directionSum % 2 == 1) {
				node.type = NodeType.CORNER;
			}
		}
		node.setConnections(validConnections);
		node.visited = true;
		return node;
	}
	
	private void traverse(Node[][] nodes, Node node) {
		Node[] connections = findConnections(nodes, node.getX(), node.getY());
		node = reclassifyNode(node, connections);
		for (Node connection: node.getConnections()) {
			if (connection != null && !connection.visited) {
				traverse(nodes, connection);
			}
		}
	}
	
	private Node[] findConnections(Node[][] nodes, int x, int y) {
		final int width = maze.getWidth();
		final int height = maze.getHeight();
		Node[] connections = new Node[4];
		if (y > 0) {
			for (int row = y-1; row >= 0; row--) {
				Node node = nodes[row][x];
				if (node != null) {
					if (node.isWall()) {
						break;
					} else {
						connections[0] = node;
						break;
					}
				}
			}
		}
		if (x < width-1) {
			for (int column = x+1; column < width; column++) {
				Node node = nodes[y][column];
				if (node != null) {
					if (node.isWall()) {
						break;
					} else {
						connections[1] = node;
						break;
					}
				}
			}
		}
		if (y < height-1) {
			for (int row = y+1; row < height; row++) {
				Node node = nodes[row][x];
				if (node != null) {
					if (node.isWall()) {
						break;
					} else {
						connections[2] = node;
						break;
					}
				}
			}
		}
		if (x > 0) {
			for (int column = x-1; column >= 0; column--) {
				Node node = nodes[y][column];
				if (node != null) {
					if (node.isWall()) {
						break;
					} else {
						connections[3] = node;
						break;
					}
				}
			}
		}
		return connections;
	}
	
	private boolean compareMazes(Node[][] maze1, Node[][] maze2) {
		final int width1 = maze1[0].length;
		final int height1 = maze1.length;
		final int width2 = maze2[0].length;
		final int height2 = maze2.length;
		
		if (width1 == width2 && height1 == height2) {
			for (int x = 0; x < height1; x++) {
				for (int y = 0; y < width1; y++) {
					if (maze1[x][y] != null && maze2[x][y] != null && maze1[x][y].type != maze2[x][y].type) {
						return false;
					}
				}
			}
		} else {
			return false;
		}
		return true;
	}
	
	private Node[][] shallowCopy(Node[][] nodes) {
		Node[][] copy = new Node[nodes.length][nodes[0].length];
		for (int x = 0; x < nodes.length; x++) {
			for (int y = 0; y < nodes[0].length; y++) {
				if (nodes[x][y] != null) {
					copy[x][y] = new Node(nodes[x][y].type, x, y, nodes[x][y].getConnections());
				}
			}
		}
		return copy;
	}
	
	public BufferedImage solve() {
		Node[][] nodes = parse();
		
		final int width = maze.getWidth();
		final int height = maze.getHeight();
		
		boolean changed = true;
		boolean finished = false;
		
		Node start = new Node(NodeType.START, 0, 0, null);
		for (int x = 0; x < height; x++) {
			for (int y = 0; y < width; y++) {
				if (nodes[x][y] != null && nodes[x][y].isStart()) {
					start = nodes[x][y];
					break;
				}
			}
		}
		
		do {
			finished = !changed;
			Node[][] oldNodes = shallowCopy(nodes);
			for (int x = 0; x < height; x++) {
				for (int y = 0; y < width; y++) {
					if (nodes[x][y] != null) {
						nodes[x][y].visited = false;
					}
				}
			}
			if (!finished) {
				traverse(nodes, start);
				changed = !compareMazes(nodes, oldNodes);
			}
		} while (!finished);
		
		ArrayList<Node> path = calculatePath(start);
		return createSolvedImage(path);
	}
	
	private ArrayList<Node> calculatePath(Node node) {
		node.visited = true;
		ArrayList<Node> path = new ArrayList<Node>();
		path.add(node);
		if (!node.isEnd()) {
			for (Node connection: node.getConnections()) {
				if (connection != null && !connection.isDeadEnd() && !connection.visited) {
					path.addAll(calculatePath(connection));
					return path;
				}
			}
			return null;
		} else {
			return path;
		}
	}
	
	private BufferedImage createSolvedImage(ArrayList<Node> path) {
		if (path == null) {
			return null;
		}
		
		final int width = maze.getWidth();
		final int height = maze.getHeight();
		
		BufferedImage solvedMaze = maze;
		final int alpha = 255;
		final int red = 255;
		final int green = 0;
		final int blue = 0;
		for (int i = 0; i < path.size()-1; i++) {
			Node node = path.get(i);
			Node next = path.get(i+1);
			
			int minX, maxX, minY, maxY;
			
			if (node.getX() < next.getX()) {
				minX = node.getX();
				maxX = next.getX();
			} else {
				minX = next.getX();
				maxX = node.getX();
			}
			if (node.getY() < next.getY()) {
				minY = node.getY();
				maxY = next.getY();
			} else {
				minY = next.getY();
				maxY = node.getY();
			}
			
			for (int x = minX; x <= maxX; x++) {
				for (int y = minY; y <= maxY; y++) {
					int rgb = solvedMaze.getRGB(x, y);
					if (rgb != GREEN && rgb != BLUE) {
						solvedMaze.setRGB(x, y, (alpha << 24) | (red << 16) | (green << 8) | blue);
					}
				}
			}
		}
		
		return solvedMaze;
	}
}

public class MazeReducer {
	
	public static void main(String[] args) {
		if (args.length == 2) {
			String input = args[0];
			String output = args[1];
			try {
				File inputFile = new File(input);
				Maze maze = new Maze(ImageIO.read(inputFile));
				BufferedImage solved = maze.solve();
				if (solved != null) {
					File outputFile = new File(output);
					try {
						ImageIO.write(solved, "png", outputFile);
					} catch (IOException e) {
						System.out.println("Could not write to output file");
						System.exit(3);
					}
				} else {
					System.out.println("Could not solve maze");
				}
			} catch (IOException e) {
				System.out.println("Could not read input file");
				System.exit(2);
			}
		} else {
			System.out.println("Two inputs required: [input file] [output file]");
			System.exit(1);
		}
		
	}
}
